--
-- Return a string JSON description of the user settings on the Motion Service.
-- Show the active location, the current ractor, and any system preferences that
-- we may want to view.
--
-- @file    json/settings.lua
-- @version 2.5
--
local base = _G

local settings = {
  ['location'] = {},
  ['ractor'] = {},
  ['system'] = {
    ['auto_export'] = 0,
    ['auto_export_path'] = '{path}/take.fbx',
    ['auto_track_take'] = 1,
    ['export_type'] = 'fbx',
    ['floor_plane'] = 1,
    ['rest_pose'] = 0,
    ['use_calibration_db'] = 1
  },
  ['export_types'] = {},
  ['imu_channel_masks'] = {
    {['key'] = 0, ['name'] = 'Default'},
    -- A, M, G, p
    {['key'] = 16805888, ['name'] = 'Raw'},
    -- Gq, Gdq, la, em, p
    {['key'] = 33030163, ['name'] = 'Quat'},
    -- Raw and Quat
    {['key'] = 16805888 + 33030163, ['name'] = 'Raw-Quat'}
  },
  ['version'] = {}
}

if 'Windows' == node._PLATFORM_NAME then
  settings['system']['reference_video'] = 0
end

-- For the remote Bus or USB tethering to a computer. Show the data acquisition
-- and internal IMU update rate settings.
if node._BUS or node.system.get_default("_shadow") then
  settings['system']['imu_channel_mask'] = 0
  settings['system']['imu_time_step'] = 0.01

  settings['sample_rate'] = {
    {['key'] = 50, ['name'] = '50'},
    {['key'] = 100, ['name'] = '100'},
    {['key'] = 200, ['name'] = '200'},
    {['key'] = 400, ['name'] = '400'},
    {['key'] = 500, ['name'] = '500'},
    {['key'] = 800, ['name'] = '800'},
    {['key'] = 1000, ['name'] = '1000'}
  }

  settings['shadow_conf'] = {
    {['key'] = '', ['name'] = 'Default'},
    {['key'] = 'v3', ['name'] = 'Shadow v3'},
    {['key'] = 'v2', ['name'] = 'Shadow v2'}
    --{['key'] = 'ue4', ['name'] = 'UE4 Mannequin'}
  }

  settings['system']['shadow_conf'] = ''
end

local location_to_table = function(obj)
  return {
    ['key'] = obj['key'],
    ['name'] = obj['name'],
    ['active'] = obj['active'],
    ['canonical_address'] = obj['canonical_address'],
    ['latitude'] = obj['latitude'],
    ['longitude'] = obj['longitude'],
    ['elevation'] = obj['elevation']
  }
end

local ractor_to_table = function(obj)
  return {
    ['name'] = obj['name'],
    ['active'] = obj['active'],
    ['key'] = obj['key'],
    ['measurement'] = {
      ['unit'] = 'cm',
      ['height'] = obj.height,
      ['arm'] = obj.arm,
      ['leg'] = obj.leg
    }
  }
end

-- Location.
local list = node.system.get_location_list()
if list then
  for _, item in ipairs(list:list()) do
    table.insert(settings.location, location_to_table(item))
  end
else
  local preference = node.system.get_preference()
  if preference then
    table.insert(settings.location, location_to_table(preference.location))
  end
end

-- Ractor.
local list = node.system.get_ractor_list()
if list then
  for _, item in ipairs(list:list()) do
    table.insert(settings.ractor, ractor_to_table(item))
  end
end

-- System preferences.
for key, value in pairs(settings.system) do
  sv = node.system.get_default(key)
  if "" ~= sv then
    settings.system[key] = sv

    -- Cast to number format.
    if base.type(1) == base.type(value) then
      settings.system[key] = base.tonumber(settings.system[key])
    end
  end
end

-- List of available export types.
local container = node.system.export_type()
if container then
  for _,value in base.pairs(container:list()) do
    table.insert(settings.export_types, {
      ['key'] = value['id'],
      ['name'] = value['description']
    })
  end
end

-- Version
local version = {
  ['version'] = node._VERSION,
  ['name'] = node._PRODUCT_NAME
}

if node._BUS then
  version['bus'] = node._BUS
end

settings.version = version

return node.json.encode(settings)
