<?xml version="1.0"?>
<!--
  Transform an XML remote procedure call definition into a Lua command string
  that returns the XML string result.
  
  In this instance, transform:
  <methodCall>... -> node.xml.command_to_xml(...)
  Where the command prints:
  node.xml.command_to_xml(...) -> <methodResponse>...
  
  Implements basic "XML-RPC" interface to the Motion Service scripting system.
  Very straight forward, except we need to escape string quotes to create the
  Lua chunk.

  @file    static/style/transform/methodCall.xsl
  @version 2.0
-->
<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform">
 <xsl:output method="text" encoding="ascii"/>

 <xsl:param name="json"/>
 
 <xsl:template match="*"/>

 <xsl:template match="/">
  <xsl:apply-templates select="methodCall"/>
 </xsl:template>

 <xsl:template match="methodCall">
  <xsl:variable name="module">
   <xsl:choose>
    <xsl:when test="$json">
     <xsl:text>json</xsl:text>
    </xsl:when>
    <xsl:otherwise>
     <xsl:text>xml</xsl:text>
    </xsl:otherwise>
   </xsl:choose>
  </xsl:variable>

  <xsl:text>return node.</xsl:text>
  <xsl:value-of select="$module"/>
  <xsl:text>.command_to_</xsl:text>
  <xsl:value-of select="$module"/>
  <xsl:text>(</xsl:text>
  <xsl:apply-templates select="methodName"/>
  <xsl:apply-templates select="params"/>
  <xsl:text>)</xsl:text>
 </xsl:template>

 <xsl:template match="methodName">
  <xsl:call-template name="escape_string">
   <xsl:with-param name="string">
    <xsl:value-of select="text()"/>
   </xsl:with-param>
  </xsl:call-template>
 </xsl:template>

 <xsl:template match="params">
  <xsl:for-each select="param">
   <xsl:text>,</xsl:text>
   <xsl:apply-templates select="value"/>
  </xsl:for-each>
 </xsl:template>

 <xsl:template match="value">
  <xsl:apply-templates mode="quote"/>
 </xsl:template>

 <!-- String values are quoted and escaped. -->
 <xsl:template match="value|string" mode="quote">
  <xsl:call-template name="escape_string">
   <xsl:with-param name="string">
    <xsl:value-of select="text()"/>
   </xsl:with-param>
  </xsl:call-template>
 </xsl:template>

 <!-- Numeric values should not be quoted. -->
 <xsl:template match="i4|int|double|number" mode="quote">
  <xsl:apply-templates/>
 </xsl:template>
 
 <!-- Convert boolean 0,1 to false,true. -->
 <xsl:template match="boolean" mode="quote">
  <xsl:choose>
   <xsl:when test="text() = '1'">
    <xsl:text>true</xsl:text>
   </xsl:when>
   <xsl:otherwise>
    <xsl:text>false</xsl:text>
   </xsl:otherwise>
  </xsl:choose>
 </xsl:template> 

 <!--
   Escape \ and ' characters in input "string" such that we can
   make a quoted string literal out of it.
 -->
 <xsl:template name="escape_string">
  <xsl:param name="string"/>

  <xsl:text>'</xsl:text>
  <xsl:call-template name="string_replace">
   <xsl:with-param name="string">
    <xsl:call-template name="string_replace">
     <xsl:with-param name="string">
      <xsl:value-of select="$string"/>
     </xsl:with-param>
     <xsl:with-param name="match">
      <xsl:text>\</xsl:text>
     </xsl:with-param>
     <xsl:with-param name="replace">
      <xsl:text>\\</xsl:text>
     </xsl:with-param>
    </xsl:call-template>
   </xsl:with-param>
   <xsl:with-param name="match">
    <xsl:text>'</xsl:text>
   </xsl:with-param>
   <xsl:with-param name="replace">
    <xsl:text>\'</xsl:text>
   </xsl:with-param>
  </xsl:call-template>
  <xsl:text>'</xsl:text>
 </xsl:template>

 <!--
   Replace all occurences of "match" in "string" with "replace".
   
   For example,
   string  = "Hello World, Hello"
   match   = "Hello"
   replace = "Goodbye"
   
   Result in,
   "Goodbye World, Goodbye"
 -->
 <xsl:template name="string_replace">
  <xsl:param name="string"/>
  <xsl:param name="match"/>
  <xsl:param name="replace"/>

  <xsl:choose>
   <xsl:when test="string-length($match)&gt;0 and contains($string, $match)">
    <xsl:value-of select="substring-before($string, $match)"/>
    <xsl:value-of select="$replace"/>
    <xsl:call-template name="string_replace">
     <xsl:with-param name="string" select="substring(substring-after($string, $match), string-length($match))"/>
     <xsl:with-param name="match" select="$match"/>
     <xsl:with-param name="replace" select="$replace"/>
    </xsl:call-template>
   </xsl:when>
   <xsl:otherwise>
    <xsl:value-of select="$string"/>
   </xsl:otherwise>
  </xsl:choose>
 </xsl:template> 

</xsl:stylesheet>