<?xml version="1.0"?>
<!--
 @file   static/xsl/t/node_to_mel.xsl
 @brief  Generate a Maya MEL command sequence to create a skeleton and Motion
         Capture device channel attachments. Requires the mDevice.mel library
         file to operate.
-->
<xsl:stylesheet
 xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
 version="1.0">

 <!-- Plain text output -->
 <xsl:output
  method="text"
  encoding="ascii"/>

 <!-- Strip all whitespace. -->
 <xsl:strip-space elements="*"/>

 <!-- Squash all nodes by default. -->
 <xsl:template match="*"/>

 <xsl:template match="/">
   <xsl:text>/**
  Automatically generated by the "node_to_mel.xsl" style sheet. Input is a
  Motion Service Configuration file formatted as XML. The output is a Maya MEL
  command set to create the skeleton, effector locators, and preview channel
  associations.

  Example usage (in MEL):

  // Create joint hierarchy.
  mDevice_create_skeleton();
  // Create locators for each of the joint FK/IK effectors.
  mDevice_create_marker();

  // Preview real-time data. Uses the Maya mocap device with associations
  // defined to the skeleton and locator set set we just loaded.
  mDevice_device(1);
*/

// Version string of the mDevice library. Also makes sure that the library
// is loaded and ready to be used.
string $mDevice_name = mDevice();

</xsl:text>

   <xsl:text>global proc string mDevice_group_name()
{
  return "</xsl:text>
   <xsl:apply-templates select="/node/@id"/>
   <xsl:text>";
}

</xsl:text>

   <xsl:text>global proc string mDevice_create_skeleton()
{
  global int $MDEVICE_IMPORT_MASS;

  string $group = mDevice_group_name();
  select -clear;
  if (objExists($group)) {
    error("Object named \"" + $group + "\" already exists");
  }
</xsl:text>

   <xsl:text>  group -empty -world -name $group;
</xsl:text>

  <xsl:apply-templates select="node/node" mode="skeleton"/>

  <xsl:text>
  return $group;
}
</xsl:text>

   <xsl:text>
global proc string mDevice_create_marker()
{
  string $marker_name = mDevice_marker_name();
  string $parent = mDevice_group_name();
  string $group = $parent + "|" + $marker_name;

  select -clear;
  if (objExists($group)) {
    error("Object named \"" + $group + "\" already exists");
  }
</xsl:text>

   <xsl:text>  group -empty -parent $parent -name $marker_name;
</xsl:text>

  <xsl:apply-templates select="node/node" mode="marker"/>

  <xsl:text>
  return $group;
}

global proc string mDevice_create_character()
{
  string $parent = mDevice_group_name();
  string $group = "TakeCharacter";

  if (!objExists($parent)) {
    error("Object named \"" + $parent + "\" does not exist");
  }

  if (objExists($group)) {
    error("Object named \"" + $group + "\" already exists");
  }

  select -replace -hierarchy $parent;

  $group = `character
    -name $group
    -excludeScale
    -excludeVisibility`;

  select -clear;

  return $group;
}

global proc mDevice_rest_pose()
{
  string $group = mDevice_group_name();</xsl:text>
  <xsl:apply-templates select="node/node/node[starts-with(@id, 'Hips')]" mode="rest_pose"/>
  <xsl:text>
}
</xsl:text>

  <xsl:apply-templates select="node" mode="mocap_list"/>
  <xsl:apply-templates select="node" mode="key_list"/>
  <xsl:apply-templates select="node" mode="marker_list"/>

  <xsl:text>
global proc string[] mDevice_device_attribute_list()
{
  string $deviceAttributeList[3] = {
    "rotateXYZ:X", "rotateXYZ:Y", "rotateXYZ:Z"
  };

  return $deviceAttributeList;
}

global proc string[] mDevice_object_attribute_list()
{
  string $objectAttributeList[3] = {
    "rotateX", "rotateY", "rotateZ"
  };

  return $objectAttributeList;
}
</xsl:text>

 </xsl:template>

 <!-- Match all non-root node entries. Skeleton generator mode. -->
 <xsl:template match="node" mode="skeleton">
  <!-- Select the parent node. -->
  <xsl:text>
  {
    select -replace ($group</xsl:text>

  <xsl:if test="count(ancestor::node) &gt; 1">
   <xsl:text> + "</xsl:text>

   <xsl:for-each select="ancestor::node[position() != last()]">
    <xsl:text>|</xsl:text>
    <xsl:value-of select="@id"/>
   </xsl:for-each>
   <xsl:text>"</xsl:text>
  </xsl:if>
  <xsl:text>);
</xsl:text>

  <xsl:choose>
   <xsl:when test="count(ancestor::node) = 1">
    <!-- Create empty group for the body container node. -->
    <xsl:text>    group -empty -parent `ls -selection -long -type transform` -name "</xsl:text>
    <xsl:value-of select="@id"/>
    <xsl:text>";
</xsl:text>
   </xsl:when>
   <xsl:otherwise>

    <!-- Create MEL "joint" statement for this node. -->
    <xsl:text>    string $name = `joint -radius 3</xsl:text>

    <xsl:if test="string-length(@rotate) > 0">
     <xsl:text> -orientation </xsl:text>
     <xsl:value-of select="translate(@rotate,',',' ')"/>
    </xsl:if>

    <xsl:if test="string-length(@translate) > 0">
     <xsl:text> -relative -position </xsl:text>
     <xsl:value-of select="translate(@translate,',',' ')"/>
    </xsl:if>

    <xsl:text> -name "</xsl:text>
    <xsl:value-of select="@id"/>
    <xsl:text>"`;
</xsl:text>

    <xsl:if test="count(ancestor::node) = 2">
     <xsl:text>    setAttr ($name + ".type") 1;
</xsl:text>
    </xsl:if>

     <xsl:if test="@marker and @mass">
       <xsl:text>
    if ($MDEVICE_IMPORT_MASS) {
      string $locator_name[] = `spaceLocator -name ($name + "_COM")`;
      setAttr ($locator_name[0] + ".translate") </xsl:text>
       <xsl:value-of select="substring-after(substring-after(substring-after(substring-after(@marker, ' '), ' '), ' '), ' ')"/>
       <xsl:text>;
      addAttr -attributeType double -defaultValue 0 -longName mass $locator_name[0];
      setAttr ($locator_name[0] + ".mass") </xsl:text>
       <xsl:value-of select="@mass"/>
       <xsl:text>;
      parent -relative $locator_name[0] $name;
    }
</xsl:text>
     </xsl:if>

   </xsl:otherwise>
  </xsl:choose>

  <xsl:text>  }
</xsl:text>

  <xsl:apply-templates select="node" mode="skeleton"/>

 </xsl:template>

  <!-- Match all non-root node entries. Skeleton generator mode. -->
  <xsl:template match="node" mode="marker">
    <xsl:text>
  {
    string $locator_name[] = `spaceLocator -name ("</xsl:text>
    <xsl:apply-templates select="@id"/>
    <xsl:text>_" + $marker_name)`;
    addAttr
      -attributeType double -defaultValue 0
      -shortName w -longName weight -keyable true $locator_name[0];
    parent $locator_name[0] $group;
  }
</xsl:text>

    <xsl:apply-templates select="node" mode="marker"/>
  </xsl:template>

 <xsl:template match="node" mode="mocap_list">
  <xsl:text>
global proc string[] mDevice_device_channel_list()
{
  string $deviceChannelList[] = {
</xsl:text>

  <xsl:for-each select=".//node">
   <xsl:text>    "</xsl:text>
   <xsl:value-of select="@id"/>
   <xsl:text>"</xsl:text>
   <xsl:if test="position() != last()">
    <xsl:text>,
</xsl:text>
   </xsl:if>
  </xsl:for-each>

  <xsl:text>
  };
  return $deviceChannelList;
}
</xsl:text>

  <xsl:text>
global proc string[] mDevice_object_channel_list()
{
</xsl:text>
  <xsl:text>  string $group = mDevice_group_name();
  if (!objExists($group)) {
    string $list[] = `ls -selection -type transform`;
    if (1 == size($list)) {
      $group = $list[0];
    }
  }
  if (!objExists($group)) {
    $group = "";
  } else {
    $group = $group + "|";
  }
</xsl:text>

  <xsl:text>
  string $objectChannelList[] = {
</xsl:text>

  <xsl:for-each select=".//node">

   <xsl:text>    ($group + </xsl:text>
   <xsl:apply-templates select="." mode="dag_path"/>
   <xsl:text>)</xsl:text>

   <xsl:if test="position() != last()">
    <xsl:text>,
</xsl:text>
   </xsl:if>
  </xsl:for-each>

  <xsl:text>
  };
  return $objectChannelList;
}
</xsl:text>

 </xsl:template>

 <xsl:template match="node" mode="key_list">
  <xsl:text>
global proc int[] mDevice_device_key_list()
{
  int $deviceKeyList[] = {
</xsl:text>

  <xsl:for-each select=".//node">
   <xsl:variable name="id" select="@id"/>
   <xsl:for-each select="/node//node">
    <xsl:sort select="@id"/>
    <xsl:if test="@id=$id">
     <xsl:text>    </xsl:text>
     <xsl:value-of select="position()"/>
    </xsl:if>
   </xsl:for-each>
   <xsl:if test="position()!=last()">
    <xsl:text>,
</xsl:text>
   </xsl:if>
  </xsl:for-each>
  <xsl:text>
  };
  return $deviceKeyList;
}
</xsl:text>
 </xsl:template>

 <xsl:template match="node" mode="marker_list">
  <xsl:text>
global proc int mDevice_copy_marker_rotate(string $marker, string $name, float $rotate[])
{
  if (objExists($name)) {
    if (!objExists($marker)) {
      spaceLocator -name $marker;
    }

    delete -constraints $marker;
    setAttr ($marker + ".rotate") $rotate[0] $rotate[1] $rotate[2];
    orientConstraint -maintainOffset $marker $name;
  } else {
    warning "Skipping marker for node named $name";
  }

  return 0;
};

global proc string mDevice_copy_marker()
{
  string $group = "Marker";
  if (!objExists($group)) {
    $group = "";
  } else {
    $group = $group + "|";
  }
</xsl:text>

  <xsl:for-each select=".//node[not(@active) and count(ancestor::node) &gt; 1 and @marker]">
   <xsl:variable name="joint">
    <xsl:apply-templates select="." mode="dag_path"/>
   </xsl:variable>
   <xsl:variable name="marker">
    <xsl:text>$group + "</xsl:text>
    <xsl:value-of select="@id"/>
    <xsl:text>_Marker"</xsl:text>
   </xsl:variable>

   <xsl:text>
  mDevice_copy_marker_rotate((</xsl:text>
   <xsl:value-of select="$marker"/>
   <xsl:text>), (</xsl:text>
   <xsl:value-of select="$joint"/>
   <xsl:text>), {</xsl:text>
   <xsl:value-of select="substring-before(@marker, ' ')"/>
   <xsl:text>, </xsl:text>
   <xsl:value-of select="substring-before(substring-after(@marker, ' '), ' ')"/>
   <xsl:text>, </xsl:text>
   <xsl:value-of select="substring-before(substring-after(substring-after(@marker, ' '), ' '), ' ')"/>
   <xsl:text>, </xsl:text>
   <xsl:value-of select="substring-before(substring-after(substring-after(substring-after(@marker, ' '), ' '), ' '), ' ')"/>
   <xsl:text>});</xsl:text>

  </xsl:for-each>
  <xsl:text>

  return $group;
}
</xsl:text>
 </xsl:template>

 <xsl:template match="node" mode="rest_pose">
  <xsl:text>
  {
    string $object = ($group + "|</xsl:text>

  <xsl:for-each select="ancestor::node[position() != last()]">
   <xsl:value-of select="@id"/>
   <xsl:text>|</xsl:text>
  </xsl:for-each>

  <xsl:value-of select="@id"/>
  <xsl:text>");

    setAttr ($object + ".translate") </xsl:text>
  <xsl:choose>
   <xsl:when test="string-length(@translate) > 0">
    <xsl:value-of select="translate(@translate,',',' ')"/>
   </xsl:when>
   <xsl:otherwise>
    <xsl:text>0 0 0</xsl:text>
   </xsl:otherwise>
  </xsl:choose>

<xsl:text>;
    select $object;
    select -hi;
    makeIdentity -apply false -t 0 -r 1 -s 0;
    select -clear;
  }</xsl:text>
 </xsl:template>

 <xsl:template match="node/@id[.='default']">
  <!-- Reserved name in Maya, "default". -->
  <xsl:text>ConfigurationGroup</xsl:text>
 </xsl:template>

 <xsl:template match="node" mode="dag_path">
  <xsl:text>"</xsl:text>
  <xsl:if test="count(ancestor::node) > 1">
   <xsl:for-each select="ancestor::node[position() != last()]">
    <xsl:value-of select="@id"/>
    <xsl:text>|</xsl:text>
   </xsl:for-each>
  </xsl:if>
  <xsl:value-of select="@id"/>
  <xsl:text>"</xsl:text>
 </xsl:template>

</xsl:stylesheet>
